/*
 * Copyright (C) 2005-2007 Andre Noll <maan@systemlinux.org>
 *
 *     This program is free software; you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation; either version 2 of the License, or
 *     (at your option) any later version.
 *
 *     This program is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public License
 *     along with this program; if not, write to the Free Software
 *     Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111, USA.
 */

/** \file audiod.c the paraslash's audio daemon */

#include "para.h"

#include "audiod.cmdline.h"
#include "list.h"
#include "sched.h"
#include "recv.h"
#include "filter.h"
#include "grab_client.cmdline.h"
#include "grab_client.h"
#include "client.cmdline.h"
#include "client.h"
#include "audiod.h"
#include "net.h"
#include "daemon.h"
#include "string.h"
#include "fd.h"
#include "write.h"
#include "write_common.h"
#include "error.h"
#include "signal.h"

/** define the array of error lists needed by para_audiod */
INIT_AUDIOD_ERRLISTS;
/** define the array containing all supported audio formats */
const char *audio_formats[] = {AUDIOD_AUDIO_FORMAT_ARRAY NULL};

/** defines how to handle one supported audio format */
struct audio_format_info {
	/** pointer to the receiver for this audio format */
	struct receiver *receiver;
	/** the receiver configuration */
	void *receiver_conf;
	/** the number of filters that should be activated for this audio format */
	unsigned int num_filters;
	/** pointer to the array of filters to be activated */
	struct filter **filters;
	/** pointer to the array of filter configurations */
	void **filter_conf;
	/** the number of filters that should be activated for this audio format */
	unsigned int num_writers;
	/** pointer to the array of writers to be activated */
	struct writer **writers;
	/** pointer to the array of writer configurations */
	void **writer_conf;
	/** do not start receiver/filters/writer before this time */
	struct timeval restart_barrier;
};

/**
 * para_audiod uses \p MAX_STREAM_SLOTS different slots, each of which may
 * be associated with a receiver/filter/writer triple. This array holds all
 * information on the status of these slots.
 *
 * \sa struct slot_info
 * */
struct slot_info slot[MAX_STREAM_SLOTS];


/**
 * the current mode of operation of which can be changed by the on/off/cycle
 * commands. It is either, AUDIOD_OFF, AUDIOD_ON or AUDIOD_STANDBY.
 */
int audiod_status = AUDIOD_ON;

/**
 * the gengetopt args_info struct that holds information on all command line
 * arguments
 */
struct audiod_args_info conf;

static char *socket_name;
static FILE *logfile;
static struct audio_format_info afi[NUM_AUDIO_FORMATS];

static struct signal_task signal_task_struct, *sig_task = &signal_task_struct;

static struct status_task status_task_struct;

/**
 * the task that calls the status command of para_server
 *
 * \sa struct status_task
 */
struct status_task *stat_task = &status_task_struct;
static struct timeval initial_delay_barrier;

/**
 * the task for handling audiod commands
 *
 * \sa struct task, struct sched
 */
struct command_task {
	/** the local listening socket */
	int fd;
	/** the associated task structure */
	struct task task;
};

/**
 * task for signal handling
 */
struct signal_task {
	/** the signal pipe */
	int fd;
	/** the number of the most recent signal */
	int signum;
	/** the associated task structure */
	struct task task;
};

/** iterate over all supported audio formats */
#define FOR_EACH_AUDIO_FORMAT(af) for (af = 0; af < NUM_AUDIO_FORMATS; af++)

/**
 * get the audio format number
 * \param name the name of the audio format
 *
 * \return The audio format number on success, -E_UNSUPPORTED_AUDIO_FORMAT if
 * \a name is not a supported audio format.
 */
int get_audio_format_num(char *name)
{
	int i;
	FOR_EACH_AUDIO_FORMAT(i)
		if (!strcmp(name, audio_formats[i]))
			return i;
	return -E_UNSUPPORTED_AUDIO_FORMAT;
}

/**
 * the log function of para_audiod
 *
 * \param ll loglevel
 * \param fmt the format string
 */
void para_log(int ll, const char* fmt,...)
{
	va_list argp;
	FILE *outfd;
	struct tm *tm;
	time_t t1;
	char str[MAXLINE] = "";
	static char *hostname;

	if (ll < conf.loglevel_arg)
		return;
	if (!hostname)
		hostname = para_hostname();
	outfd = logfile? logfile : stderr;
	time(&t1);
	tm = localtime(&t1);
	strftime(str, MAXLINE, "%b %d %H:%M:%S", tm);
	fprintf(outfd, "%s %s ", str, hostname);
	if (conf.loglevel_arg <= INFO)
		fprintf(outfd, "%i ", ll);
	va_start(argp, fmt);
	vfprintf(outfd, fmt, argp);
	va_end(argp);
}

static char *configfile_exists(void)
{
	static char *config_file;

	if (!config_file) {
		char *home = para_homedir();
		config_file = make_message("%s/.paraslash/audiod.conf", home);
		free(home);
	}
	return file_exists(config_file)? config_file : NULL;
}

static void setup_signal_handling(void)
{
	sig_task->fd = para_signal_init();
	PARA_INFO_LOG("signal pipe: fd %d\n", sig_task->fd);
	para_install_sighandler(SIGINT);
	para_install_sighandler(SIGTERM);
	para_install_sighandler(SIGHUP);
	signal(SIGPIPE, SIG_IGN);
}

static void clear_slot(int slot_num)
{
	struct slot_info *s = &slot[slot_num];

	PARA_INFO_LOG("clearing slot %d\n", slot_num);
	memset(s, 0, sizeof(struct slot_info));
	s->format = -1;
}

static void close_receiver(int slot_num)
{
	struct slot_info *s = &slot[slot_num];
	struct audio_format_info *a;

	if (s->format < 0 || !s->receiver_node)
		return;
	a = &afi[s->format];
	PARA_NOTICE_LOG("closing %s receiver in slot %d (eof = %d)\n",
		audio_formats[s->format] , slot_num, s->receiver_node->eof);
	a->receiver->close(s->receiver_node);
	free(s->receiver_node);
	s->receiver_node = NULL;
}

static void kill_all_decoders(void)
{
	int i;

	FOR_EACH_SLOT(i) {
		struct slot_info *s = &slot[i];
		if (s->wng && !s->wng->eof) {
			PARA_INFO_LOG("unregistering writer node group in slot %d\n",
				i);
			wng_unregister(s->wng);
			s->wng->eof = 1;
		}
		if (s->fc && !s->fc->eof) {
			PARA_INFO_LOG("unregistering filter chain in slot %d\n", i);
			unregister_task(&s->fc->task);
			s->fc->eof = 1;
		}
		if (s->receiver_node && !s->receiver_node->eof) {
			PARA_INFO_LOG("unregistering receiver_node in slot %d\n", i);
			unregister_task(&s->receiver_node->task);
			s->receiver_node->eof = 1;
		}
	}
}

static int get_empty_slot(void)
{
	int i;
	struct slot_info *s;

	FOR_EACH_SLOT(i) {
		s = &slot[i];
		if (s->format < 0) {
			clear_slot(i);
			return i;
		}
		if (s->wng || s->receiver_node || s->fc)
			continue;
		clear_slot(i);
		return i;
	}
	return -E_NO_MORE_SLOTS;
}

/**
 * get the number of filters
 *
 * \param audio_format_num the number identifying the audio format
 *
 * \return the number of filters for the given audio format
 *
 * \sa struct filter;
 */
int num_filters(int audio_format_num)
{
	return afi[audio_format_num].num_filters;
}

static void filter_event_handler(struct task *t)
{
	PARA_NOTICE_LOG("%s\n", PARA_STRERROR(-t->ret));
	struct filter_chain *fc = t->private_data;
	fc->eof = 1;
	unregister_task(t);
}

static void open_filters(int slot_num)
{
	struct slot_info *s = &slot[slot_num];
	struct audio_format_info *a = &afi[s->format];
	int nf = a->num_filters;
	int i;

	s->fc = NULL;
	if (!nf)
		return;
	PARA_INFO_LOG("opening %s filters\n", audio_formats[s->format]);
	s->fc = para_calloc(sizeof(struct filter_chain));
	INIT_LIST_HEAD(&s->fc->filters);
	s->fc->inbuf = s->receiver_node->buf;
	s->fc->in_loaded = &s->receiver_node->loaded;
	s->fc->input_eof = &s->receiver_node->eof;
	s->fc->task.pre_select = filter_pre_select;
	s->fc->task.event_handler = filter_event_handler;
	s->fc->task.private_data = s->fc;
	s->fc->eof = 0;

	s->receiver_node->output_eof = &s->fc->eof;
	sprintf(s->fc->task.status, "filter chain");
	for (i = 0; i < nf; i++) {
		struct filter_node *fn = para_calloc(sizeof(struct filter_node));
		fn->conf = a->filter_conf[i];
		fn->fc = s->fc;
		fn->filter = a->filters[i];
		INIT_LIST_HEAD(&fn->callbacks);
		list_add_tail(&fn->node, &s->fc->filters);
		fn->filter->open(fn);
		PARA_NOTICE_LOG("%s filter %d/%d (%s) started in slot %d\n",
			audio_formats[s->format], i + 1,  nf,
			fn->filter->name, slot_num);
		s->fc->outbuf = fn->buf;
		s->fc->out_loaded = &fn->loaded;
	}
	register_task(&s->fc->task);
}

static void wng_event_handler(struct task *t)
{
	struct writer_node_group *wng = t->private_data;

	PARA_INFO_LOG("%s\n", PARA_STRERROR(-t->ret));
	wng->eof = 1;
	wng_unregister(wng);
}

static void open_writers(int slot_num)
{
	int ret, i;
	struct slot_info *s = &slot[slot_num];
	struct audio_format_info *a = &afi[s->format];

	PARA_INFO_LOG("opening %s writers\n", audio_formats[s->format]);
	if (!a->num_writers)
		s->wng = setup_default_wng();
	else
		s->wng = wng_new(a->num_writers);
	if (s->fc) {
		s->wng->buf = s->fc->outbuf;
		s->wng->loaded = s->fc->out_loaded;
		s->wng->input_eof = &s->fc->eof;
		s->wng->channels = &s->fc->channels;
		s->wng->samplerate = &s->fc->samplerate;
		s->fc->output_eof = &s->wng->eof;
		PARA_INFO_LOG("samplerate: %d\n", *s->wng->samplerate);
	} else {
		s->wng->buf = s->receiver_node->buf;
		s->wng->loaded = &s->receiver_node->loaded;
		s->wng->input_eof = &s->receiver_node->eof;
	}
	s->wng->task.event_handler = wng_event_handler;
	for (i = 0; i < a->num_writers; i++) {
		s->wng->writer_nodes[i].conf = a->writer_conf[i];
		s->wng->writer_nodes[i].writer = a->writers[i];
	}
	ret = wng_open(s->wng);
	if (ret < 0) {
		PARA_ERROR_LOG("%s\n", PARA_STRERROR(-ret));
		return;
	}
	s->wstime = *now;
	activate_inactive_grab_clients(slot_num, s->format, &s->fc->filters);
}

static void rn_event_handler(struct task *t)
{
	struct receiver_node *rn = t->private_data;
	const struct timeval restart_delay = {0, 10 * 1000};
	int i;

	PARA_NOTICE_LOG("%s\n", PARA_STRERROR(-t->ret));
	unregister_task(t);
	rn->eof = 1;
	/* set restart barrier */
	FOR_EACH_SLOT(i) {
		if (slot[i].receiver_node != rn)
			continue;
		tv_add(now, &restart_delay, &afi[slot[i].format].restart_barrier);
	}
}

static int open_receiver(int format)
{
	struct audio_format_info *a = &afi[format];
	struct slot_info *s;
	int ret, slot_num;
	struct receiver_node *rn;
	const struct timeval restart_delay = {1, 0};

	ret = get_empty_slot();
	if (ret < 0)
		goto err;
	slot_num = ret;
	s = &slot[slot_num];
	s->format = format;
	s->receiver_node = para_calloc(sizeof(struct receiver_node));
	rn = s->receiver_node;
	rn->receiver = a->receiver;
	rn->conf = a->receiver_conf;
	ret = a->receiver->open(s->receiver_node);
	if (ret < 0) {
		free(s->receiver_node);
		s->receiver_node = NULL;
		goto err;
	}
	PARA_NOTICE_LOG("started %s: %s receiver in slot %d\n",
		audio_formats[s->format], a->receiver->name, slot_num);
	rn->task.private_data = s->receiver_node;
	rn->task.pre_select = a->receiver->pre_select;
	rn->task.post_select = a->receiver->post_select;
	rn->task.event_handler = rn_event_handler;
	sprintf(rn->task.status, "%s receiver node", rn->receiver->name);
	register_task(&rn->task);
	return 1;
err:
	PARA_ERROR_LOG("%s\n", PARA_STRERROR(-ret));
	tv_add(now, &restart_delay, &afi[format].restart_barrier);
	return ret;
}

static int receiver_running(int format)
{
	int i;

	FOR_EACH_SLOT(i) {
		struct slot_info *s = &slot[i];
		if (s->format == format && s->receiver_node
				&& !s->receiver_node->eof)
			return 1;
	}
	return 0;
}

static int open_current_receiver(struct sched *s)
{
	int i;
	struct timeval diff;
	char *audio_format = stat_task->stat_item_values[SI_FORMAT];

	if (!audio_format || !stat_task->pcd)
		return 0;
	i = get_audio_format_num(audio_format + strlen(
		status_item_list[SI_FORMAT]) + 1);
	if (i < 0)
		return 0;
	if (receiver_running(i))
		return 0;
	if (tv_diff(now, &afi[i].restart_barrier, &diff) < 0) {
		s->timeout = diff;
		return 0;
	}
	return open_receiver(i) < 0? 0 : 1;
}

static unsigned compute_time_diff(const struct timeval *status_time)
{
	struct timeval tmp, diff;
	static unsigned count;
	int sign, sa_time_diff_sign = stat_task->sa_time_diff_sign;
	const struct timeval max_deviation = {0, 500 * 1000};
	const int time_smooth = 5;

	if (!status_time)
		return count;
	sign = tv_diff(status_time, now, &diff);
//		PARA_NOTICE_LOG("%s: sign = %i, sa_time_diff_sign = %i\n", __func__,
//			sign, sa_time_diff_sign);
	if (!count) {
		sa_time_diff_sign = sign;
		stat_task->sa_time_diff = diff;
		count++;
		goto out;
	}
	if (count > 5) {
		int s = tv_diff(&diff, &stat_task->sa_time_diff, &tmp);
		if (tv_diff(&max_deviation, &tmp, NULL) < 0)
			PARA_WARNING_LOG("time diff jump: %lims\n",
				s * tv2ms(&tmp));
	}
	count++;
	sa_time_diff_sign = tv_convex_combination(
		sa_time_diff_sign * time_smooth, &stat_task->sa_time_diff,
		count > 10? sign : sign * time_smooth, &diff,
		&tmp);
	stat_task->sa_time_diff = tmp;
	PARA_INFO_LOG("time diff (cur/avg): %s%lums/%s%lums\n",
		sign > 0? "+" : "-",
		tv2ms(&diff),
		sa_time_diff_sign ? "+" : "-",
		tv2ms(&stat_task->sa_time_diff)
	);
out:
	stat_task->sa_time_diff_sign = sa_time_diff_sign;
	return count;
}

static void check_stat_line(char *line)
{
	int itemnum;
	size_t ilen = 0;
	long unsigned sec, usec;
	char *tmp;

//	PARA_INFO_LOG("line: %s\n", line);
	if (!line)
		return;
	itemnum = stat_line_valid(line);
	if (itemnum < 0) {
		PARA_WARNING_LOG("invalid status line: %s\n", line);
		return;
	}
	if (stat_task->clock_diff_count && itemnum != SI_CURRENT_TIME)
		return;
	tmp = make_message("%s\n", line);
	stat_client_write(tmp, itemnum);
	free(tmp);
	free(stat_task->stat_item_values[itemnum]);
	stat_task->stat_item_values[itemnum] = para_strdup(line);
	ilen = strlen(status_item_list[itemnum]);
	switch (itemnum) {
	case SI_STATUS:
		stat_task->playing = strstr(line, "playing")? 1 : 0;
		break;
	case SI_OFFSET:
		stat_task->offset_seconds = atoi(line + ilen + 1);
		break;
	case SI_LENGTH:
		stat_task->length_seconds = atoi(line + ilen + 1);
		break;
	case SI_STREAM_START:
		if (sscanf(line + ilen + 1, "%lu.%lu", &sec, &usec) == 2) {
			struct timeval a_start, delay;
			delay.tv_sec = conf.stream_delay_arg / 1000;
			delay.tv_usec = (conf.stream_delay_arg % 1000) * 1000;
			stat_task->server_stream_start.tv_sec = sec;
			stat_task->server_stream_start.tv_usec = usec;
			if (compute_time_diff(NULL) > 2) {
				if (stat_task->sa_time_diff_sign < 0)
					tv_add(&stat_task->server_stream_start,
						&stat_task->sa_time_diff, &a_start);
				else
					tv_diff(&stat_task->server_stream_start,
						&stat_task->sa_time_diff, &a_start);
				tv_add(&a_start, &delay, &initial_delay_barrier);
			}
		}
		break;
	case SI_CURRENT_TIME:
		if (sscanf(line + ilen + 1, "%lu.%lu", &sec, &usec) == 2) {
			struct timeval tv = {sec, usec};
			compute_time_diff(&tv);
		}
		if (stat_task->clock_diff_count)
			stat_task->clock_diff_count--;
		break;
	}
}

static void try_to_close_slot(int slot_num)
{
	struct slot_info *s = &slot[slot_num];

	if (s->format < 0)
		return;
	if (s->receiver_node && !s->receiver_node->eof)
		return;
	if (s->fc && !s->fc->eof)
		return;
	if (s->wng && !s->wng->eof)
		return;
	PARA_INFO_LOG("closing slot %d \n", slot_num);
	wng_close(s->wng);
	close_filters(s->fc);
	free(s->fc);
	close_receiver(slot_num);
	clear_slot(slot_num);
}

/*
 * Check if any receivers/filters/writers need to be started and do so if
 * neccessary.  Since the pre_select function didn't have a chance yet to put
 * file descriptors into the fd sets given by s, make the upcoming select()
 * return immediately to avoid a long timeout in case we started something.
 */
static void audiod_pre_select(struct sched *s, __a_unused struct task *t)
{
	int i;
	struct timeval min_delay = {0, 1};

	t->ret = 1;
	if (audiod_status != AUDIOD_ON || !stat_task->playing)
		return kill_all_decoders();
	if (open_current_receiver(s))
		s->timeout = min_delay;
	FOR_EACH_SLOT(i) {
		struct slot_info *sl = &slot[i];
		struct audio_format_info *a;
		struct timeval diff;

		if (sl->format < 0)
			continue;
		a = &afi[sl->format];
		if (!sl->receiver_node)
			continue;
		if (!a->num_filters) {
			if (sl->receiver_node->loaded && !sl->wng) {
				open_writers(i);
				s->timeout = min_delay;
			}
			continue;
		}
		if (sl->receiver_node->loaded && !sl->fc) {
			open_filters(i);
			s->timeout = min_delay;
			continue;
		}
		if (!sl->fc || !*sl->fc->out_loaded || sl->wng)
			continue;
		if (tv_diff(now, &initial_delay_barrier, &diff) > 0) {
			open_writers(i);
			s->timeout = min_delay;
			continue;
		}
		PARA_INFO_LOG("inital delay: %lu ms left\n", tv2ms(&diff));
		if (tv_diff(&s->timeout, &diff, NULL) > 0) {
			s->timeout = diff;
		}
	}
}

static void audiod_post_select(__a_unused struct sched *s,
	__a_unused struct task *t)
{
	int i;

	t->ret = 1;
	FOR_EACH_SLOT(i)
		try_to_close_slot(i);
}

static void init_audiod_task(struct task *t)
{
	t->pre_select = audiod_pre_select;
	t->post_select = audiod_post_select;
	t->event_handler = NULL;
	t->private_data = t;
	sprintf(t->status, "audiod task");
}

static int parse_stream_command(const char *txt, char **cmd)
{
	char *p = strchr(txt, ':');
	int i;

	if (!p)
		return -E_MISSING_COLON;
	p++;
	FOR_EACH_AUDIO_FORMAT(i) {
		if (strncmp(txt, audio_formats[i], strlen(audio_formats[i])))
			continue;
		*cmd = p;
		return i;
	}
	return -E_UNSUPPORTED_AUDIO_FORMAT;
}

static int add_filter(int format, char *cmdline)
{
	struct audio_format_info *a = &afi[format];
	int filter_num, nf = a->num_filters;

	filter_num = check_filter_arg(cmdline, &a->filter_conf[nf]);
	if (filter_num < 0)
		return filter_num;
	a->filters[nf] = &filters[filter_num];
	a->num_filters++;
	PARA_INFO_LOG("%s filter %d: %s\n", audio_formats[format], nf + 1,
		a->filters[nf]->name);
	return filter_num;
}

static int init_writers(void)
{
	int i, ret, nw;
	char *cmd;
	struct audio_format_info *a;

	init_supported_writers();
	nw = PARA_MAX(1, conf.writer_given);
	PARA_INFO_LOG("maximal number of writers: %d\n", nw);
	FOR_EACH_AUDIO_FORMAT(i) {
		a = &afi[i];
		a->writer_conf = para_malloc(nw * sizeof(void *));
		a->writers = para_malloc(nw * sizeof(struct writer *));
		a->num_writers = 0;
	}
	for (i = 0; i < conf.writer_given; i++) {
		void *wconf;
		int writer_num;
		ret = parse_stream_command(conf.writer_arg[i], &cmd);
		if (ret < 0)
			goto out;
		a = &afi[ret];
		nw = a->num_writers;
		wconf = check_writer_arg(cmd, &writer_num);
		if (!wconf) {
			ret = writer_num;
			goto out;
		}
		a->writers[nw] = &writers[writer_num];
		a->writer_conf[nw] = wconf;
		PARA_INFO_LOG("%s writer #%d: %s\n", audio_formats[ret],
			nw, writer_names[writer_num]);
		a->num_writers++;
	}
	ret = 1;
out:
	return ret;
}

static int init_receivers(void)
{
	int i, ret, receiver_num;
	char *cmd = NULL;
	struct audio_format_info *a;

	for (i = 0; receivers[i].name; i++) {
		PARA_INFO_LOG("initializing %s receiver\n", receivers[i].name);
		receivers[i].init(&receivers[i]);
	}
	for (i = conf.receiver_given - 1; i >= 0; i--) {
		char *arg = conf.receiver_arg[i];
		char *recv_arg = strchr(arg, ':');
		ret = -E_MISSING_COLON;
		if (!recv_arg)
			goto out;
		*recv_arg = '\0';
		recv_arg++;
		ret = get_audio_format_num(arg);
		if (ret < 0)
			goto out;
		afi[ret].receiver_conf = check_receiver_arg(recv_arg, &receiver_num);
		if (!afi[ret].receiver_conf) {
			ret = -E_RECV_SYNTAX;
			goto out;
		}
		afi[ret].receiver = &receivers[receiver_num];
	}
	/* use the first available receiver with no arguments
	 * for those audio formats for which no receiver
	 * was specified
	 */
	cmd = para_strdup(receivers[0].name);
	FOR_EACH_AUDIO_FORMAT(i) {
		a = &afi[i];
		if (a->receiver_conf)
			continue;
		a->receiver_conf = check_receiver_arg(cmd, &receiver_num);
		if (!a->receiver_conf)
			return -E_RECV_SYNTAX;
		a->receiver = &receivers[receiver_num];
	}
	ret = 1;
out:
	free(cmd);
	return ret;
}

static int init_default_filters(void)
{
	int i, ret = 1;

	FOR_EACH_AUDIO_FORMAT(i) {
		struct audio_format_info *a = &afi[i];
		char *tmp;
		int j;

		if (a->num_filters)
			continue; /* no default -- nothing to to */
		/* add "dec" to audio format name */
		tmp = make_message("%sdec", audio_formats[i]);
		for (j = 0; filters[j].name; j++)
			if (!strcmp(tmp, filters[j].name))
				break;
		free(tmp);
		ret = -E_UNSUPPORTED_FILTER;
		if (!filters[j].name)
			goto out;
		tmp = para_strdup(filters[j].name);
		ret = add_filter(i, tmp);
		free(tmp);
		if (ret < 0)
			goto out;
		PARA_INFO_LOG("%s -> default filter: %s\n", audio_formats[i],
			filters[j].name);
	}
out:
	return ret;
}

static int init_filters(void)
{
	int i, ret, nf;

	filter_init(filters);
	nf = PARA_MAX(1,  conf.filter_given);
	PARA_INFO_LOG("maximal number of filters: %d\n", nf);
	FOR_EACH_AUDIO_FORMAT(i) {
		afi[i].filter_conf = para_malloc(nf * sizeof(void *));
		afi[i].filters = para_malloc(nf * sizeof(struct filter *));
	}
	if (!conf.no_default_filters_given)
		return init_default_filters();
	for (i = 0; i < conf.filter_given; i++) {
		char *arg = conf.filter_arg[i];
		char *filter_name = strchr(arg, ':');
		ret = -E_MISSING_COLON;
		if (!filter_name)
			goto out;
		*filter_name = '\0';
		filter_name++;
		ret = get_audio_format_num(arg);
		if (ret < 0)
			goto out;
		ret = add_filter(ret, filter_name);
		if (ret < 0)
			goto out;
	}
	ret = init_default_filters(); /* use default values for the rest */
out:
	return ret;
}

static int init_stream_io(void)
{
	int ret;

	ret = init_writers();
	if (ret < 0)
		return ret;
	ret = init_receivers();
	if (ret < 0)
		return ret;
	ret = init_filters();
	if (ret < 0)
		return ret;
	return 1;
}

static int audiod_get_socket(void)
{
	struct sockaddr_un unix_addr;
	int fd;

	if (conf.socket_given)
		socket_name = para_strdup(conf.socket_arg);
	else {
		char *hn = para_hostname();
		socket_name = make_message("/var/paraslash/audiod_socket.%s",
			hn);
		free(hn);
	}
	PARA_NOTICE_LOG("local socket: %s\n", socket_name);
	if (conf.force_given)
		unlink(socket_name);
	fd = create_pf_socket(socket_name, &unix_addr,
			S_IRUSR | S_IWUSR | S_IRGRP | S_IWGRP | S_IWOTH);
	if (fd < 0) {
		PARA_EMERG_LOG("%s", "can not connect to socket\n");
		exit(EXIT_FAILURE); /* do not unlink socket */
	}
	if (listen(fd , 5) < 0) {
		PARA_EMERG_LOG("%s", "can not listen on socket\n");
		exit(EXIT_FAILURE); /* do not unlink socket */
	}
	mark_fd_nonblock(fd);
	return fd;
}

static void signal_event_handler(struct task *t)
{
	struct signal_task *st = t->private_data;

	switch (st->signum) {
	case SIGINT:
	case SIGTERM:
	case SIGHUP:
		PARA_EMERG_LOG("terminating on signal %d\n", st->signum);
		clean_exit(EXIT_FAILURE, "caught deadly signal");
	}
}

static void signal_pre_select(struct sched *s, struct task *t)
{
	struct signal_task *st = t->private_data;
	t->ret = 1;
	para_fd_set(st->fd, &s->rfds, &s->max_fileno);
}

static void signal_post_select(struct sched *s, struct task *t)
{
	struct signal_task *st = t->private_data;
	t->ret = 1;
	if (!FD_ISSET(st->fd, &s->rfds))
		return;
	t->ret = -E_SIGNAL_CAUGHT;
	st->signum = para_next_signal();
}

static void signal_setup_default(struct signal_task *st)
{
	st->task.pre_select = signal_pre_select;
	st->task.post_select = signal_post_select;
	st->task.private_data = st;
	sprintf(st->task.status, "signal task");
}

static void command_pre_select(struct sched *s, struct task *t)
{
	struct command_task *ct = t->private_data;
	t->ret = 1;
	para_fd_set(ct->fd, &s->rfds, &s->max_fileno);

}

static void command_post_select(struct sched *s, struct task *t)
{
	int ret;
	struct command_task *ct = t->private_data;

	t->ret = 1; /* always successful */
	audiod_status_dump();
	if (!FD_ISSET(ct->fd, &s->rfds))
		return;
	ret = handle_connect(ct->fd);
	if (ret < 0)
		PARA_ERROR_LOG("%s\n", PARA_STRERROR(-ret));
}

static void init_command_task(struct command_task *ct)
{
	ct->task.pre_select = command_pre_select;
	ct->task.post_select = command_post_select;
	ct->task.event_handler = NULL;
	ct->task.private_data = ct;
	ct->fd = audiod_get_socket(); /* doesn't return on errors */
	sprintf(ct->task.status, "command task");
}

static void close_stat_pipe(void)
{
	int i;

	if (!stat_task->pcd)
		return;
	client_close(stat_task->pcd);
	stat_task->pcd = NULL;
	for (i = 0; i < NUM_STAT_ITEMS; i++) {
		free(stat_task->stat_item_values[i]);
		stat_task->stat_item_values[i] = NULL;
	}
	dump_empty_status();
	stat_task->length_seconds = 0;
	stat_task->offset_seconds = 0;
	audiod_status_dump();
	stat_task->playing = 0;
	stat_task->stat_item_values[SI_STATUS_BAR] = make_message(
		"%s:no connection to para_server\n",
		status_item_list[SI_STATUS_BAR]);
	stat_client_write(stat_task->stat_item_values[SI_STATUS_BAR],
		SI_STATUS_BAR);
	if (stat_task->clock_diff_count) {
		stat_task->clock_diff_barrier.tv_sec = now->tv_sec + 1;
		stat_task->clock_diff_barrier.tv_usec = now->tv_usec;
	}
}

/**
 * close the connection to para_server and exit
 *
 * \param status the exit status which is passed to exit(3)
 * \param msg the log message
 *
 * Log \a msg with loglevel \p EMERG, close the connection to para_server if
 * open, and call \p exit(status). \a status should be either EXIT_SUCCESS or
 * EXIT_FAILURE.
 *
 * \sa exit(3)
 */
void __noreturn clean_exit(int status, const char *msg)
{
	PARA_EMERG_LOG("%s\n", msg);
	if (socket_name)
		unlink(socket_name);
	close_stat_pipe();
	exit(status);
}


static void client_task_event_handler(__a_unused struct task *t)
{
	int i;
	struct timeval delay = {1, 0};

	if (t->ret == -E_HANDSHAKE_COMPLETE)
		return;
	unregister_task(t);
	close_stat_pipe();
	if (t->ret != -E_SERVER_EOF)
		stat_task->clock_diff_count = conf.clock_diff_count_arg;
	/* avoid busy loop if server is down */
	tv_add(now, &delay, &stat_task->restart_barrier);
	FOR_EACH_AUDIO_FORMAT(i)
		afi[i].restart_barrier = stat_task->restart_barrier;
}

static void status_pre_select(struct sched *s, struct task *t)
{
	struct status_task *st = t->private_data;
	int ret;

	t->ret = 1; /* always successful */
	if (st->pcd || audiod_status == AUDIOD_OFF)
		return;
	if (!st->clock_diff_count && tv_diff(now, &st->restart_barrier, NULL)
			< 0)
		return;
	if (st->clock_diff_count) {
		char *argv[] = {"audiod", "stat", "1", NULL};
		int argc = 3;
		if (tv_diff(now, &st->clock_diff_barrier, NULL) < 0)
			return;
		PARA_INFO_LOG("clock diff count: %d\n", st->clock_diff_count);
		ret = client_open(argc, argv, &st->pcd);

	} else {
		char *argv[] = {"audiod", "stat", NULL};
		int argc = 2;
		ret = client_open(argc, argv, &st->pcd);
	}
	if (ret < 0)
		return;
	st->pcd->task.event_handler = client_task_event_handler;
	s->timeout.tv_sec = 0;
	s->timeout.tv_usec = 1;
}

static void status_post_select(__a_unused struct sched *s, struct task *t)
{
	struct status_task *st = t->private_data;
	unsigned bytes_left;

	t->ret = 1;
	if (!st->pcd || st->pcd->status != CL_RECEIVING)
		return;
	if (st->pcd && audiod_status == AUDIOD_OFF) {
		unregister_task(&st->pcd->task);
		close_stat_pipe();
		st->clock_diff_count = conf.clock_diff_count_arg;
		return;
	}
	bytes_left = for_each_line(st->pcd->buf, st->pcd->loaded,
		&check_stat_line);
	if (st->pcd->loaded != bytes_left) {
		st->last_status_read = *now;
		st->pcd->loaded = bytes_left;
	} else {
		struct timeval diff;
		tv_diff(now, &st->last_status_read, &diff);
		if (diff.tv_sec > 61)
			close_stat_pipe();
	}
}

static void init_status_task(struct status_task *st)
{
	memset(st, 0, sizeof(struct status_task));
	st->task.pre_select = status_pre_select;
	st->task.post_select = status_post_select;
	st->task.private_data = st;
	st->sa_time_diff_sign = 1;
	st->clock_diff_count = conf.clock_diff_count_arg;
	sprintf(st->task.status, "status task");
}

static void set_initial_status(void)
{
	audiod_status = AUDIOD_ON;
	if (!conf.mode_given)
		return;
	if (!strcmp(conf.mode_arg, "sb")) {
		audiod_status = AUDIOD_STANDBY;
		return;
	}
	if (!strcmp(conf.mode_arg, "off")) {
		audiod_status = AUDIOD_OFF;
		return;
	}
	if (strcmp(conf.mode_arg, "on"))
		PARA_WARNING_LOG("%s", "invalid mode\n");
}

/**
 * the main function of para_audiod
 *
 * \param argc usual argument count
 * \param argv usual argument vector
 *
 * \return EXIT_SUCCESS or EXIT_FAILURE
 *
 * \sa para_audiod(1)
 * */
int main(int argc, char *argv[])
{
	char *cf;
	int ret, i;
	struct sched s;
	struct command_task command_task_struct, *cmd_task = &command_task_struct;
	struct task audiod_task_struct, *audiod_task = &audiod_task_struct;

	valid_fd_012();
	audiod_cmdline_parser(argc, argv, &conf);
	HANDLE_VERSION_FLAG("audiod", conf);
	para_drop_privileges(conf.user_arg, conf.group_arg);
	cf = configfile_exists();
	if (cf) {
		if (audiod_cmdline_parser_configfile(cf, &conf, 0, 0, 0)) {
			PARA_EMERG_LOG("%s", "parse error in config file\n");
			exit(EXIT_FAILURE);
		}
	}
	if (conf.logfile_given)
		logfile = open_log(conf.logfile_arg);
	log_welcome("para_audiod", conf.loglevel_arg);
	i = init_stream_io();
	if (i < 0) {
		PARA_EMERG_LOG("init stream io error: %s\n", PARA_STRERROR(-i));
		exit(EXIT_FAILURE);
	}
	server_uptime(UPTIME_SET);
	set_initial_status();
	FOR_EACH_SLOT(i)
		clear_slot(i);
	init_grabbing();
	setup_signal_handling();
	signal_setup_default(sig_task);
	sig_task->task.event_handler = signal_event_handler;

	init_status_task(stat_task);
	init_command_task(cmd_task);
	init_audiod_task(audiod_task);

	if (conf.daemon_given)
		daemon_init();

	register_task(&sig_task->task);
	register_task(&cmd_task->task);
	register_task(&stat_task->task);
	register_task(audiod_task);
	s.default_timeout.tv_sec = 0;
	s.default_timeout.tv_usec = 99 * 1000;
	ret = sched(&s);

	PARA_EMERG_LOG("%s\n", PARA_STRERROR(-ret));
	return EXIT_FAILURE;
}
